# -*- coding: utf-8 -*-
"""
6.100A Spring 2023: Recitation 4

Lecture 7:
- Data abstraction
- Classes

Lecture 8:
- Inheritance
- Subtypes
- Extended classes example
- Class attributes
"""

from enum import Enum

def instantiation_example():
    class Car(object):
        """
        Represents a specific make/model vehicle
        to be analyzed across several factors.
        """

        def __init__(self, make, model, mpg):
            """
            args:
                make    <str>   : company name of vehicle
                model   <str>   : which vehicle name
                mpg     <float> : fuel efficiency in miles per gallon
            """
            self.make = make
            self.model = model
            self.mpg = mpg
        
        def set_mpg(self, nmpg):
            """
            miles per gallon setter
            args:
                nmpg    <float> : new miles per gallon efficiency
            """
            self.mpg = nmpg
        
        def get_mpg(self):
            """
            miles per gallon getter
            """
            return self.mpg

    c1 = Car("Pontiac", "Torrent", 2)
    print(c1)

def dunder_methods():
    from car import Car
    c1 = Car("Pontiac", "Torrent", 2)
    print(c1)

def instance_methods():
    from car import Car
    c1 = Car("Pontiac", "Torrent", 2)
    print(c1)

    c1.set_mpg(10)
    print(c1)

def instance_attributes():
    from car import Car
    c1 = Car("Chevy", "Camaro", 22)
    print(c1)

    # our "Car" type is mutable
    c1.make = "Ford"
    print(c1)

def returning_aliases_vs_copies():
    from student import Student
    matt = Student("Matt")
    matt.add_grade(10)
    matt.add_grade(20)
    matt.add_grade(30)

    john = Student("John")
    john.add_grade(100)
    john.add_grade(98)
    john.add_grade(93)

    try:
        matt.add_grade("A") # assertion error
    except AssertionError as e:
        print(f"AssertionError: {e}")

    matt_grades = matt.get_grades_alias()
    matt_grades[0] = "A"
    print(matt)

    john_grades = john.get_grades_copy()
    john_grades[0] = "A"
    print(john)

def class_attributes():
    from license import LicensePlate
    plate1 = LicensePlate("ABC123")
    print(plate1)
    plate2 = LicensePlate("987XYZ")
    print(plate2)

    try:
        plate3 = LicensePlate("ABC123")
    except ValueError as e:
        print(f"ValueError: {e}")

def enum_example():
    # enums list out what values are valid
    class Direction(Enum):
        NORTH = 0
        EAST = 1
        SOUTH = 2
        WEST = 3
    
    direction = Direction.NORTH
    print(direction)

def inheritance_example():
    from car import Car
    class Sedan(Car):
        def __init__(self, make, model, mpg, doors):
            # initialize our Car methods and attributes
            super().__init__(make, model, mpg)
            self.doors = doors

    s = Sedan("Honda", "Civic", 33, 4)
    print(s)
    print(s.get_mpg())

def isinstance_vs_type():
    from car import Car
    from sedan import Sedan

    car = Car("Ford", "F150", 20)
    sedan = Sedan("Honda", "Civic", 33, 4)

    print("type(car) == Car", type(car) == Car)
    print("type(sedan) == Sedan", type(sedan) == Sedan)
    print("type(car) == Sedan", type(car) == Sedan)
    print("type(sedan) == Car", type(sedan) == Car)

    print()

    print("isinstance(car, Car)", isinstance(car, Car))
    print("isinstance(sedan, Sedan)", isinstance(sedan, Sedan))
    print("isinstance(car, Sedan)", isinstance(car, Sedan))
    print("isinstance(sedan, Car)", isinstance(sedan, Car))

def polymorphism_example():
    from car import Car
    from sedan import Sedan

    cars = []
    cars.append(Car("Ford", "F150", 20))
    cars.append(Sedan("Honda", "Civic", 33, 4))

    print("using polymorphism:")

    for c in cars:
        print(c)

    print("\nnot using polymorphism:")

    for c in cars:
        print(Car.__str__(c))

if __name__ == "__main__":
    # instantiation_example()
    # dunder_methods()
    # instance_methods()
    # instance_attributes()
    # returning_aliases_vs_copies()
    # class_attributes()
    enum_example()
    # inheritance_example()
    # isinstance_vs_type()
    # polymorphism_example()
    pass